#! /usr/bin/perl
#
# @(#)$Id: testbuild-lcg.pl,v 1.1 2015/06/10 15:49:50 pmacvsdg Exp $
#
# This script builds the EGI.eu CA distribution, based on the IGTF 
# release as published on the web site, and then amended as per EGI.eu
# Council policy.
# As a result, this meta-RPM encodes dependendies on *individual* CAs,
# and no longer on the IGTF AP bundles, since the EGI.eu Council
# decided to no endorse individual CAs within those bundles.
#
# Using the EGI.eu distribution is therefore at your own risk -- do
# do bother the IGTF with either error messages or politics, please
#
package main;

use strict;
use Getopt::Long qw(:config no_ignore_case bundling);
use POSIX;
use File::Temp qw( tempfile );
use File::Copy;
use Archive::Tar;
eval { require LWP or die; }; $@ and die "Please install libwww-perl (LWP)\n";

my $year = strftime("%Y",localtime);
my $today = strftime("%d/%m/%Y",localtime);
my $textheader = <<EOF;
# Copyright (c) Stichting FOM/Nikhef for LCG, $year
# This work is licensed under the Creative Commons Attribution
# 3.0 License. To view a copy of this license, visit 
#   http://creativecommons.org/licenses/by/3.0/
# Generated by $0
#
# Warning: this template encodes the LCG policy on endorsed CAs. Your
# site or national policies, and agreements with other customers, may have 
# added CAs to the set that you are advised to install. 
# *** Use of this template is AT YOUR OWN RISK ***
EOF

use vars qw/ $srcurl $output $verb $help $version $release $specin /;
use vars qw/ $profiles $exclusions $additions /;
use vars qw/ %options $target /;
use vars qw/ $uprelease /;
use vars qw/ @otherspecfiles /;
my $rpmsignargs;

local %options = (
    "srcurl=s" => \$srcurl,
    "o|output=s" => \$output,
    "version=s" => \$version,
    "r|release=s" => \$release,
    "uprelease=s" => \$uprelease,
    "t|target=s" => \$target,
    "s|specfile=s" => \$specin,
    "otherspecfiles=s@" => \@otherspecfiles,
    "excl|exclusions=s" => \$exclusions,
    "incl|additions|inclusions=s" => \$additions,
    "ap|profiles=s" => \$profiles,
    "v|verbose+" => \$verb,
    "h|help+" => \$help,
  );

sub help(%) {
  my %opt = @_;
  my $length=8+length($0);
  print "Usage: $0";
  foreach my $v ( keys %opt ) {
    if ($length+length($v)+2 > 77 ) { print "\n    "; $length=5; }
    print " [$v]" and $length+=(length($v)+3);
  }
  print <<EOF;

Put together the EGI policy distribution based on the web publication of
the IGTF Distribution version. Write the result to the non-existent or
empty directory <output>.
EOF
  return 1;
}

&GetOptions(%options) or &help(%options) and exit(1);
$help and &help(%options) and exit(0);

# sensible defaults
$release ||= 1;
$uprelease ||= 1;
$srcurl ||= "http://dist.eugridpma.info/distribution/igtf/current-new/";
$specin ||= 'ca-policy-egi.spec.in';
$profiles ||= '(classic|mics|slcs)';
$exclusions ||= '(^$)';
$additions ||= '(^cilogon-basic$)';
$target ||= 'groep@lxplus.cern.ch:/afs/cern.ch/user/g/groep/www/lcg-ca/distribution/';
$#otherspecfiles >= 0 or @otherspecfiles = qw/ ca-policy-lcg.spec.cin lcg-CA.spec.cin /;

-f "$ENV{HOME}/.rpmmacros.egi" and 
  $rpmsignargs = "--macros=$ENV{HOME}/.rpmmacros.egi";

chomp(my $srcrpmdir=`rpm --eval %_srcrpmdir`);
chomp(my $sourcedir=`rpm --eval %_sourcedir`);

defined $version or do {
  my $ua = LWP::UserAgent->new;
  my $response = $ua->get("$srcurl/version.txt");
  if ( ! $response->is_success ) {
    die "Cannot retrieve version from $srcurl: ".$response->status_line."\n";
  }
  my $vdata = $response->content;
  $vdata =~ m/^This is version (\d+\.\d+)/ or
    die "Cannot find version number in $srcurl/version.txt\n$vdata\n";
  $version = $1;
};

$output ||= "ca-policy-lcg-$version-$release";
$output =~ s/VERSION/$version-$release/g;

$verb and print "Building for version $version release $release\n";
$verb and print "Source URL $srcurl\n";
$verb and print "Output directory $output\n";
$verb and print "Supported APs: $profiles\n";
$verb and print "Exclusions: $exclusions\n";
$verb and print "Additions: $additions\n";

-f "doc/ca-policy-lcg-readme-$version.txt" or 
  die "Cannot find compulsory doc/ca-policy-lcg-readme-$version.txt";

# retrieve list of obsoleted CAs for all profiles
# as these are the 'known bad' ones, as opposeed to ones that are
# 'just not in the EGI set'
my @obsoletedcalist;
while ($profiles =~ /(\w+)/g ) {
  my $ap = $1;
  $verb > 1 and print "Retrieveing obsoleted list for $ap\n";
  open RPM,"rpm -qp --nosignature --obsoletes ".
           "$srcurl/accredited/RPMS/ca_policy_igtf-$ap-$version-$uprelease.noarch.rpm|" 
    or die "Cannot retrieve Obsoletes data from IGTF RPM: $!\n";
  while (<RPM>) {
    chomp($_); s/^\s*//; s/\s*$//;
    /^[-_a-zA-Z0-9]+$/ and push @obsoletedcalist,$_;
  }
}
$verb and print "Upstream obsoletions included: ".$#obsoletedcalist."\n";

# retrieve list of accredited CAs from source URL
#
my @allcalist;
my @calist;
my %cainfo;
{
  my $ua = LWP::UserAgent->new;
  my $response = $ua->get("$srcurl/accredited/accredited.in");
  if ( ! $response->is_success ) {
    die "Cannot retrieve accredited list from $srcurl: ".$response->status_line."\n";
  }
  my $vdata = $response->content;
  foreach my $line ( split(/[\n\r]+/,$vdata) ) {
    $verb>2 and print "Considering $line\n";
    my ($name,$ap,$comment) = split(/\s+/,$line,3);
    push @allcalist,"accredited/RPMS/ca_$name-$version-$uprelease.noarch.rpm";
    $ap =~ /$profiles/ or next;
    $name !~ /$exclusions/ or next;
    push @calist,$name;
  }
}

# additions from unaccredited, experimental, and accredited but not proper profile
foreach my $class ( qw/ experimental unaccredited accredited / ) {
  my $ua = LWP::UserAgent->new;
  my $response = $ua->get("$srcurl/$class/RPMS/");
  if ( ! $response->is_success ) {
    die "Cannot retrieve $class list from $srcurl: ".$response->status_line."\n";
  }
  my $vdata = $response->content;
  foreach my $line ( split(/[\n\r]+/,$vdata) ) {
    $verb>2 and print "Considering for addition line $line\n";
    $line =~ m/.*<a href=\"ca_([^\"]+)-$version-[^-]+\.noarch\.rpm\">ca_.*/i or next;
    my $name = ($1);
    $verb and print "Considering $name for addition per $additions\n";
    $name =~ /$additions/ or next;
    $verb and print "Explicit addition $name added to calist\n";
    push @calist,$name;
    push @allcalist,"$class/RPMS/ca_$name-$version-$uprelease.noarch.rpm";
  }
}


$verb>1 and print "@calist\n";

#
# building RPMs for EGI and others
#
my @otherrpms = qw/ ca-policy-egi-core /;

my $specdata;
$specin and do {
  open SPEC,"<$specin" or die "Cannot open $specin: $!\n";
  while (<SPEC>) { $specdata .= $_; };
  close SPEC
};

my $descr = <<EOF;
This package encodes the policy on Approval of Certificate Authorities,
as documented in JSPG Policy Document https://edms.cern.ch/document/428038,
as adopted and amended by the EGI.eu Council and its executive bodies.
By this policy it only incorporated those trust anchors accredited by 
the IGTF under the profiles $profiles, except those for 
which the Council has decided to withdraw its endorsement, namely 
$exclusions, and the authorities added: $additions
EOF
$specdata =~ s!\@DESCRIPTION\@!$descr!gm;

my $requires;
foreach ( @calist ) {
  $requires and $requires.=", ";
  $requires .= "ca_$_ = $version";
}
my $addrequires;
foreach ( @calist ) {
  print "Testing $_ as bing an adition in $additions\n";
  $_ =~ m!$additions! or next;
  print "... YES!\n";
  $addrequires.=", ";
  $addrequires .= "ca_$_ = $version";
}
my $obsoletes;
foreach ( @obsoletedcalist ) {
  $obsoletes and $obsoletes.=", ";
  $obsoletes .= "$_";
}
$specdata =~ s!\@REQUIRES\@!$requires!gm;
$specdata =~ s!\@OBSOLETES\@!$obsoletes!gm;
$specdata =~ s!\@VERSION\@!$version!gm;
$specdata =~ s!\@RELEASE\@!$release!gm;

my ($tmpspec, $tmpspecname) = 
  tempfile("ca-policy-egi-$version.spec.XXXXXX", CLEANUP => 1);
$tmpspec or die "Cannot create temporary spec file: $!\n";

print $tmpspec  $specdata;
system("rpmbuild","-ba","--quiet","$tmpspecname");
close $tmpspec and unlink("$tmpspecname");

#
# collect RPMs
chomp(my $rpmdir = `rpm --eval '%_rpmdir'`);

-e "$output" and ! -d "$output" and die "$output: not a directory\n";
-d "$output/RPMS" and die "$output: directory not empty\n";
-d "$output" or mkdir "$output" or die "mkdir $output: $!\n";
-d "$output/meta" or mkdir "$output/meta" or die "mkdir $output/meta: $!\n";
mkdir "$output/RPMS" or die "mkdir $output/RPMS: $!\n";
mkdir "$output/tgz" or die "mkdir $output/tgz: $!\n";
mkdir "$output/repo-files" or die "mkdir $output/repo-files: $!\n";

# copy the key metadata for yum configuration
copy("doc/lcg-trustanchors.repo", "$output/repo-files/lcg-trustanchors.repo");
copy("doc/GPG-KEY-EUGridPMA-RPM-3", "$output/GPG-KEY-EUGridPMA-RPM-3");
copy("doc/LICENSE.lcg", "$output/meta/LICENSE");

#
# collect the dummy CA RPM to fix the renegotiation bug
#
my @rpmnames;
copy("bin/dummy-ca-certs-20090630-1.noarch.rpm",
     "$output/RPMS/dummy-ca-certs-20090630-1.noarch.rpm");
push @rpmnames, "$output/RPMS/dummy-ca-certs-20090630-1.noarch.rpm";

foreach my $rpmname ( @allcalist ) {
  my $targetname = $rpmname;
  $targetname =~ s/.*\///; # clean any leading paths
  my $ua = LWP::UserAgent->new;
  my $response = $ua->mirror("$srcurl/$rpmname", 
                             "$output/RPMS/$targetname");
  if ( ! $response->is_success and $response->status_line !~ /304/ ) {
    die "Cannot retrieve '$rpmname' from $srcurl: ".$response->status_line."\n";
  }

  # source distro support without dependency support
  my $tgzname = $rpmname;
  $tgzname =~ s!/RPMS/!/tgz/!gm;
  $tgzname =~ s!-[^-]+\.noarch\.rpm!\.tar\.gz!gm;
  $targetname = $tgzname;
  $targetname =~ s/.*\///; # clean any leading paths
  my $response = $ua->mirror("$srcurl/$tgzname", 
                             "$output/tgz/$targetname");
  if ( ! $response->is_success and $response->status_line !~ /304/ ) {
    die "Cannot retrieve '$tgzname' from $srcurl: ".$response->status_line."\n";
  }

  # extract the subjectDN from the TGZ package
  my $tar = Archive::Tar->new("$output/tgz/$targetname");
  $targetname  =~ m/ca_(.+)\-\d+\.\d+\.tar\.gz$/;
  my $alias  = $1;
  print "Extracting subjectDN for $alias:\n";
  my $infodata = $tar->get_content("ca_$alias-$version/$alias.info");

  $infodata =~ m/.*subjectdn\s*=\s*\"([^\"]+)\".*/m;
  my $subjectdn = $1; die "No subject DN for $alias" unless $subjectdn;
  $infodata =~ m/.*version\s*=\s*(\S+).*/m;
  my $caversion = $1; die "No CA specific version for $alias" unless $subjectdn;
  $cainfo{$alias}->{"subjectdn"} = $subjectdn;
  $cainfo{$alias}->{"version"} = $caversion;

  $verb and print "  $alias ($caversion): subjectDN = $subjectdn\n";
}

foreach my $ap ( split(/[\|\,\(\)]+/,$profiles) ) {
  $ap or next;
  my $rpmname = "ca_policy_igtf-$ap-$version-$uprelease.noarch.rpm";
  my $ua = LWP::UserAgent->new;
  my $response = $ua->mirror("$srcurl/accredited/RPMS/$rpmname", 
                             "$output/RPMS/$rpmname");
  if ( ! $response->is_success ) {
    die "Cannot retrieve $rpmname from $srcurl: ".$response->status_line."\n";
  }
}

# construct the policy info file for ca-policy-lcg, and copy rest
my $nca; my $fdata="";
$fdata .= "# policy-lcg.info - wLCG policy information file\n";
$fdata .= "# Accept IGTF Classic, MICS, and SLCS plus any wLCG specifics\n";
$fdata .= "version = $version\n";
$fdata .= "requires = "; $nca = 0 ; foreach my $ca ( @calist ) {
  $fdata .= ", \\\n  " if $nca;
  $fdata .= "$ca = ".$cainfo{$ca}->{"version"}; $nca++;
} $fdata .= "\n";
$fdata .= "subjectdn = "; $nca = 0 ; foreach my $ca ( @calist ) {
  $fdata .= ", \\\n  " if $nca;
  $fdata .= "\"".$cainfo{$ca}->{"subjectdn"}."\""; $nca++;
} $fdata .= "\n";
## CONTINUE HERE - THIS FILE NEEDS TO BE WRITTEN TO A TAR ARCH AND
## INCLUDED IN THE RP

# collect readme and documentation
copy("doc/ca-policy-lcg-readme-$version.txt","$output/meta/");
if ( -f "README.txt" ) {
  copy("README.txt","$output/");
} else {
  copy("doc/ca-policy-lcg-readme-$version.txt","$output/README.txt");
}

# build related metapackages for as long as these are allowed in the
# EGI repositories. Conversely, LCG can mirror all of the EGI repo
#
foreach my $otherspec ( @otherspecfiles ) {
  my $specdata;
  open SPEC,"<$otherspec" or die "Cannot open $otherspec: $!\n";
  while (<SPEC>) { $specdata .= $_; };
  close SPEC;
  if ( $otherspec =~ /\.cin$/ ) {
    $specdata =~ s!\@VERSION\@!$version!gm;
    $specdata =~ s!\@RELEASE\@!$release!gm;
    $specdata =~ s!\@ADDREQUIRES\@!$addrequires!gm;
  }
  my ($tmpspec, $tmpspecname) = 
    tempfile("egibuildspec-$version.spec.XXXXXX", CLEANUP => 1);
  $tmpspec or die "Cannot create temporary spec file: $!\n";
  print $tmpspec  $specdata;
  system("rpmbuild","-ba","--quiet","$tmpspecname");
  close $tmpspec and unlink("$tmpspecname");
  $specdata =~ /.*Name: ([^\n]+)\s*\n/mi;
  push @otherrpms, $1;
  $verb and print "Adding additional RPM $1\n";
}

foreach my $rpm ( @otherrpms ) {
  move ("$rpmdir/noarch/$rpm-$version-$release.noarch.rpm",
        "$output/RPMS/");
  push @rpmnames,"$output/RPMS/$rpm-$version-$release.noarch.rpm";
}
$verb and print "Signing RPM @rpmnames with $rpmsignargs\n";
system("rpm --resign -v $rpmsignargs @rpmnames");

# now generate distribution meta-data and config for sites outside an RPM

open NRMS,">$output/meta/ca-policy-lcg-$version-$release.nrms" or
  warn "Cannot write NRMS meta-data: $!\n";
defined *NRMS and do {
  my $releasenotes;
  open NOTES,"<doc/ca-policy-lcg-readme-$version.txt" or 
    die "Cannot open doc/ca-policy-lcg-readme-$version.txt: $!\n";
  while (<NOTES>) { $releasenotes .= $_; };
  close NOTES;

  print NRMS <<EOF;
<Release>
    <DistributionName>LCG Trust Anchor Distribution</DistributionName>
    <DistributionShortName>CAs</DistributionShortName>
    <SoftwareProvider>wLCG</SoftwareProvider>
    <SoftwareComponent>ca-policy-lcg</SoftwareComponent>
    <Contact>davidg\@nikhef.nl</Contact>
    <Description>Trust anchors endored by the wLCG project and EGI compatibility packages</Description>
    <Keywords>ca-policy-lcg trust CA IGTF</Keywords>
    <DocumentationLinks>https://www.eugridpma.org/</DocumentationLinks>
    <Date>$today</Date>
    <Version>$version-$release</Version>
    <SynchProtocol>http</SynchProtocol>
    <SynchURL>http://cern.ch/lcg-ca/distribution/current/</SynchURL>
    <ChangeLog>Update of the LCG Trust Anchors to IGTF release $version-$uprelease</ChangeLog>
    <ReleaseNotes>
$releasenotes
</ReleaseNotes>
</Release>
EOF
  close *NRMS;
  copy("$output/meta/ca-policy-lcg-$version-$release.nrms","$output/meta/release.xml");
};

open TPL,">$output/meta/pro_software_meta_ca_policy_lcg.tpl" or
  warn "Cannot write classic Quattor template: $!\n";
open NSTPL,">$output/meta/ca-policy-lcg.tpl" or
  warn "Cannot write namespaced Quattor template: $!\n";
open RPMLIST,">$output/meta/ca-policy-lcg.list" or
  warn "Cannot write RPM list: $!\n";
open OBSOLIST,">$output/meta/ca-policy-lcg.obsoleted" or
  warn "Cannot write obsoleted-aliases list: $!\n";

defined *TPL and print TPL <<EOF;
#
# @(#)pro_software_meta_ca_policy_lcg.tpl $version-$release
$textheader#
template pro_software_meta_ca_policy_lcg;

"/software/packages"=pkg_repl("ca-policy-lcg","$version-$release","noarch");
"/software/packages"=pkg_repl("ca_policy_igtf-classic","$version-$release","noarch");
"/software/packages"=pkg_repl("ca_policy_igtf-mics","$version-$release","noarch");
"/software/packages"=pkg_repl("ca_policy_igtf-slcs","$version-$release","noarch");
EOF

defined *NSTPL and print NSTPL <<EOF;
#
# @(#)ca-policy-lcg.tpl $version-$release
$textheader# 
template common/security/ca-policy-lcg;

"/software/packages"=pkg_repl("ca-policy-lcg","$version-$release","noarch");
"/software/packages"=pkg_repl("ca_policy_igtf-classic","$version-$release","noarch");
"/software/packages"=pkg_repl("ca_policy_igtf-mics","$version-$release","noarch");
"/software/packages"=pkg_repl("ca_policy_igtf-slcs","$version-$release","noarch");
EOF

defined *RPMLIST and print RPMLIST <<EOF;
#
# @(#)ca-policy-lcg.tpl $version-$release
$textheader# 
ca-policy-lcg-$version-$release
ca_policy_igtf-classic-$version-$release
ca_policy_igtf-mics-$version-$release
ca_policy_igtf-slcs-$version-$release
EOF

foreach my $ca ( @calist ) {
  defined *TPL and print TPL 
    "\"/software/packages\"=pkg_repl(\"ca_$ca\",".
    "\"$version-$uprelease\",\"noarch\");\n";
  defined *NSTPL and print NSTPL 
    "\"/software/packages\"=pkg_repl(\"ca_$ca\",".
    "\"$version-$uprelease\",\"noarch\");\n";
  defined *RPMLIST and print RPMLIST 
    "ca_$ca-$version-$uprelease\n";
}

defined *OBSOLIST and print OBSOLIST <<EOF;
#
# @(#)ca-policy-lcg.obsoleted $version-$release
#
# List of obsoleted trust anchors. The trust anchors identified by these
# aliases have been withdrawn by their owners or issuers because they have 
# been superceded, have expired, or have been compromised.
# 
EOF

foreach ( @obsoletedcalist ) {
  s/^ca_//;
  print OBSOLIST "$_\n";
}

defined *TPL and close *TPL;
defined *NSTPL and close *NSTPL;
defined *RPMLIST and close *RPMLIST;
defined *OBSOLIST and close *OBSOLIST;

# now collate and prepare the distribution for installation
-d "$output/RPMS" and do {
  system("createrepo -s sha $output");
  -x "/usr/bin/yum-arch" and system("yum-arch","$output");
};

$target and do {
  print "Please acknowledge distribution to target area:\n  $target\n\nOr press ^C NOW\n";
  <>;
  system("rsync","-e","ssh","-ra","$output","$target");
};


sub quoteDN($) {
  my ($s) = @_;
  $s =~ s/\%/\%25/g;
  $s =~ s/\"/\%22/g;
  return "\"$s\"";
}

